@@ -50,8 +50,8 @@
  * organization.
  */
 #define KRB5_MAJOR_RELEASE 1
-#define KRB5_MINOR_RELEASE 10
-#define KRB5_PATCHLEVEL 7
+#define KRB5_MINOR_RELEASE 11
+#define KRB5_PATCHLEVEL 0
 /* #undef KRB5_RELTAIL */
-#define KRB5_RELDATE "20131106"
-#define KRB5_RELTAG "krb5-1.10.7-final"
+#define KRB5_RELDATE "20121217"
+#define KRB5_RELTAG "krb5-1.11-final"
